import numpy as np

class Strip(object):
    def __init__(self, io, offset, length, compression):
        self.io = io
        self.offset = offset
        self.length = length
        self.compression = compression


    def read(self, dtype):
        old_position = self.io.pos()
        self.io.seek(self.offset)
        buf = self.io.read_bytes(self.length)

        if self.compression == 1:
            pass
        elif self.compression == 5:
            buf = lzw_decompress.tiff_lzw_decompress(buf)

        self.io.seek(old_position)
        return np.frombuffer(buf, dtype=dtype)


    def __repr__(self):
        msg = '<Strip offset=0x{:x} len={} compression={}>'
        return msg.format(self.offset, self.length, self.compression)



import re
import sys
import itertools
from collections import OrderedDict

import numpy as np

import bigtiff.tif_format
from bigtiff.strip import Strip
from bigtiff.memmap import memmap as my_memmap

class Image2dIterator(object):
    def __init__(self, first_ifd):
        self.next_ifd = first_ifd


    def __next__(self):
        if self.next_ifd is None:
            raise StopIteration

        ifd = self.next_ifd
        self.next_ifd = self.next_ifd.next_ifd
        return Image2d(ifd)


class Image2d(object):
    '''
    This represents one image in the TIF file.

    WARNING:
    Although this class is called Image2d it can either be a grayscale image
    dim=(X,Y,1) or a RGB image dim=(X,Y,3).
    This is due to the way TIFF saves the images in the file.
    '''
    def __init__(self, ifd):
        self.ifd = ifd

        tags = self.tags
        self.height = tags['image_length'][0]
        self.width = tags['image_width'][0]

        # self.write_image = np.ma.masked_all([height, width, self.n_channels])


    @property
    def n_channels(self):
        channels = self.tags['samples_per_pixel'][0]
        return channels


    @property
    def axes(self):
        # ImageJ storage order is TZCXY
        desc = self.tags['image_description'][0].string

        images = re.search('images=([0-9]+)', desc)
        channels = re.search('channels=([0-9]+)', desc)
        slices = re.search('slices=([0-9]+)', desc)
        frames = re.search('frames=([0-9]+)', desc)

        axes = [('X', 999, self.width), ('Y', 888, self.height)]
        if channels:
            axes.append(('C', channels.start(), channels.group(1)))
        else:
            axes.append(('C', -1, self.n_channels))
        if slices:
            axes.append(('Z', slices.start(), slices.group(1)))
        if frames:
            axes.append(('T', frames.start(), frames.group(1)))

        axes = OrderedDict((a, int(value)) for a, start, value in sorted(axes, key=lambda pair: pair[1]))
        return axes


    @property
    def shape(self):
        return (self.width, self.height, self.n_channels)


    @property
    def tags(self):
        tags = {
            'samples_per_pixel': [1],
            'sample_format': [1],
            'strip_offsets': [],
            'strip_byte_counts': [],
            'compression': [1],
            'image_width': [0],
            'rows_per_strip': [],
            'image_length': [0],
        }

        for entry in self.ifd.entries:
            if entry.is_value:
                tags[entry.tag.name] = entry.values.array
            else:
                tags[entry.tag.name] = entry.external_values.array

        return tags


    @property
    def dtype(self):
        tags = self.tags

        endian = self.ifd._root.endian
        endian = {
            endian.be: '>',
            endian.le: '<',
        }[endian]

        typ = tags['sample_format'][0]
        typ = {
             1: 'u',
             2: 'i',
             3: 'f',
        }[typ]

        length = np.ceil(tags['bits_per_sample'][0] / 8.0)
        return np.dtype('{}{}{}'.format(endian, typ, int(length)))


    @property
    def strips(self):
        '''List of Strips (data slices) for this image'''
        tags = self.tags
        strip_offsets = tags['strip_offsets']
        strip_byte_counts = tags['strip_byte_counts']

        if strip_offsets is None or strip_byte_counts is None:
            return []

        compression = tags['compression'] * len(strip_offsets)

        return [ Strip(self.ifd._io, offset, length, compr)
                 for offset, length, compr in zip(strip_offsets, strip_byte_counts, compression) ]


    def __getitem__(self, slices):
        '''
        Deprecated: Use self.memmap() instead.
        Get image data (use three indices: H, W, C).
        '''
        assert len(slices) == 3

        first_row = None
        strip_list = []
        for i, strip, strip_start, strip_stop in strip_iterator((slices[0].start or 0),
                                                                (slices[0].stop or tags['image_length'])):
            first_row = first_row or strip_start
            strip_list.append(strip.read(self.dtype_tif))

        data = np.concatenate(strip_list)
        data = data.reshape([-1, tags['image_width'][0], self.n_channels])

        slices = list(slices)
        slices[0] = slice(slices[0].start - first_row, slices[0].stop - first_row, slices[0].step)
        return data[slices[0], slices[1], slices[2]]


    def strip_iterator(self, start_row, stop_row):
        '''
        Iterate over strips that include rows from start_row to stop_row
        '''
        tags = self.tags
        strip_offsets = tags['strip_offsets']
        rows_per_strip = tags['rows_per_strip']

        if len(rows_per_strip) == len(strip_offsets) - 1:
            # last strip is missing
            rows_per_strip.append(sys.maxsize)

        if len(rows_per_strip) == 1:
            # we have only one value for all strips
            rows_per_strip = rows_per_strip * len(strip_offsets)

        start_row_per_strip = np.hstack([[0], np.cumsum(rows_per_strip[:-1])])
        end_row_per_strip = start_row_per_strip + rows_per_strip

        for i, strip, strip_start, strip_stop in enumerate(zip(self.strips, start_row_per_strip, end_row_per_strip)):
            if strip_start <= start_row:
                if strip_start <= stop_row:
                    yield i, strip, strip_start, strip_stop
                else:
                    break


    def __setitem__(self, slices, values):
        '''Deprecated: Use self.memmap() instead'''
        raise NotImplemented


    def flush(self):
        '''
        Deprecated: Use self.memmap() instead.
        Write assignments done using __setitem__() to disk.
        '''
        raise NotImplemented

        tags = self.tags
        width = tags['image_length']

        edges = np.ma.notmasked_edges(self.write_image, axis=0)
        if edges is None:
            return

        for row_start, row_stop in edges:
            first_row = None
            for i, strip, strip_start, strip_end in strip_iterator(row_start, row_stop):
                assert tags['compression'][i] == 1
                first_row = first_row or stip_start


    def memmap(self):
        '''Returns an `numpy.memmap()`-ed array of the image'''
        tags = self.tags
        H = tags['image_length'][0]
        W = tags['image_width'][0]
        C = tags['samples_per_pixel'][0]

        strips = self.strips
        pos = strips[0].offset
        for s in strips:
            if pos != s.offset:
                msg = 'Cannot memmap non-consecutive image strips (image has {} strips)'
                raise NotImplementedError(msg.format(len(strips)))

            pos += s.length

        if strips[0].compression != 1:
            raise NotImplementedError('Cannot memmap compressed images')

        array = my_memmap(strips[0].io._io, mode='r+', dtype=self.dtype,
                          shape=(H, W, C), offset=strips[0].offset)

        if array.shape[-1] == 1:
            return array[:, :, 0]
        else:
            return array



import io
import sys
import struct

import bigtiff.tif_format as tif_format

class Writer(object):
    def __init__(self, io, endian=None):
        '''io: either io.BytesIO or io.BufferedWriter'''
        endian = endian or sys.byteorder

        if endian in '<>!=@':
            self.endian = endian
        else:
            self.endian = '<' if endian.startswith('l') else '>'

        self.io = io

    def close(self):
        self.io.close()

    def __len__(self):
        if type(self.io) == io.BytesIO:
            return len(self.io.getbuffer())
        elif type(self.io) == io.BufferedWriter:
            return self.io.tell()
        else:
            raise NotImplemented

    def get_bytes(self):
        if type(self.io) == io.BytesIO:
            return self.io.getbuffer()
        else:
            raise NotImplemented

    def write_bytes(self, buf):
        self.io.write(buf)

    def write_u16(self, v):
        return self.pack('?H', v)

    def write_i16(self, v):
        return self.pack('?h', v)

    def write_u32(self, v):
        return self.pack('?I', v)

    def write_i32(self, v):
        return self.pack('?i', v)

    def write_u64(self, v):
        return self.pack('?Q', v)

    def write_i64(self, v):
        return self.pack('?q', v)

    def write_f32(self, v):
        return self.pack('?f', v)

    def write_f64(self, v):
        return self.pack('?d', v)

    def write_hole(self, length):
        if type(self.io) == io.BytesIO:
            return self.write(b'\0' * length)
        elif type(self.io) == io.BufferedWriter:
            self.io.seek(length - 1, io.SEEK_CUR)
            res = self.io.write(b'\0')
            assert res != -1
        else:
            raise NotImplemented

    def pack(self, fmt, *v):
        self.io.write(struct.pack(fmt.replace('?', self.endian), *v))

    def write_tag_value(self, typ, v, offset):
        TagType = tif_format.TifFormat.TagType

        fmt = {
            TagType.u2:      '?H',
            TagType.u4:      '?I',
            TagType.u8:      '?Q',
            TagType.u_ratio: '?II',
            TagType.s2:      '?h',
            TagType.s4:      '?i',
            TagType.s8:      '?q',
            TagType.s_ratio: '?ii',
            TagType.string:  '?{}s',
        }[typ]

        if typ == TagType.string:
            return self.pack(fmt.format(len(v)), v)
        elif typ not in (TagType.u_ratio, TagType.s_ratio):
            return self.pack(fmt, v)
        else:
            return self.pack(fmt, *v)




# Python version of LZW decompressor
# much slower than the cython version!

import numpy as np

class EoiCode(object):
    pass

class ClearCode(object):
    pass

TIFF_TABLE = [
    bytes([i]) for i in range(256)
]
TIFF_TABLE.append(ClearCode)
TIFF_TABLE.append(EoiCode)

TIFF_CODE_LEN = [9, 12]


class BitReader(object):
    def __init__(self, buf):
        self.buf = buf
        self.curr = 0
        self.read = 0

    mask = [0x00, 0x01, 0x03, 0x07, 0x0f, 0x1f, 0x3f, 0x7f]

    def readbits(self, n):
        left = n - self.read
        if left > 8:
            self.curr = (self.curr << 8) | (next(self.buf) & 0xff)
            left -= 8
        self.read = 8 - left
        next_byte = next(self.buf) & 0xff

        print(next_byte, self.read)
        print('a', (self.curr << left), '', (next_byte >> self.read))
        r = (self.curr << left) | (next_byte >> self.read)
        self.curr = next_byte & self.mask[self.read]
        return r


class LZWError(Exception):
    pass


def lzw_decompress(init_table, code_len, buf):
    min_code_len, max_code_len = code_len
    code_len = min_code_len
    table = []+init_table
    reader = BitReader(iter(buf))
    out = b''

    while True:
        code = reader.readbits(code_len)
        if code < len(table):
            value = table[code]
        else:
            value = None

        if value == EoiCode:
            break
        elif value == ClearCode:
            print('CLEAR')
            table = [] + init_table
            code_len = min_code_len

            code = reader.readbits(code_len)
            value = table[code]

            if value == EoiCode:
                break

            out += value
        else:
            if value is not None:
                print('1', old_code, code, code_len)
                out += value
                table.append(table[old_code] + value[:1])
            else:
                print('2', old_code, code, code_len)
                old_value = table[old_code]
                string = old_value + old_value[:1]
                out += string
                table.append(string)
        old_code = code

        if len(table) == 511:
            code_len = 10
        elif len(table) == 1023:
            code_len = 11
        elif len(table) == 2047:
            code_len = 12

    return out


def bitstring_to_bytes(s):
    return int(s, 2).to_bytes(len(s) // 8, byteorder='big')


if __name__ == '__main__':
    b = [256, 7, 258, 8, 8, 258, 6, 257]
    b = ('{:09b}'*len(b)).format(*b)
    b = bitstring_to_bytes(b)
    res = lzw_decompress(TIFF_TABLE, TIFF_CODE_LEN, b)

    np.testing.assert_array_equal(np.frombuffer(res, dtype='uint8'), [7,7,7,8,8,7,7,6])




import bigtiff.tif_format as tif_format

tif_format.TifFormat.URatio.__repr__ = lambda self: repr(float(self.nominator)/self.denominator)
tif_format.TifFormat.SRatio.__repr__ = lambda self: repr(float(self.nominator)/self.denominator)
tif_format.TifFormat.String.__repr__ = lambda self: repr(self.string)




import sys
import io as _io
import contextlib

import numpy as np

import bigtiff.tif_format as tif_format
from kaitaistruct import KaitaiStream

from bigtiff.image2d import Image2dIterator
from bigtiff.writer import Writer
from bigtiff import representation

class Tiff(tif_format.TifFormat):
    def __iter__(self):
        if not self.header.is_big_tiff:
            return Image2dIterator(self.header.first_ifd)
        else:
            return Image2dIterator(self.header.first_ifd_big_tiff)


    def __del__(self):
        # close tiff file when object is destroyed
        self._io.close()

    @classmethod
    def from_file(cls, filename):
        # we need write permission for memmap
        f = open(str(filename), 'br+')

        return cls(KaitaiStream(f))

    @classmethod
    def from_fd(cls, fd):
        return cls(KaitaiStream(fd))

    @classmethod
    def memmap_tcz(cls, filename):
        fd = open(str(filename), 'br+')
        slices = [s for s in cls.from_fd(fd)]

        axes = slices[0].axes.copy()
        ax_keys = list(axes.keys())
        assert 'X', 'Y' in ax_keys[-2:]  # the plane must be XY or YX

        if ax_keys[-2:] == ('Y', 'X'):
            t = lambda s: s.T
        else:
            t = lambda s: s

        samples_per_pixel = slices[0].tags['samples_per_pixel'][0]

        # we must trick numpy to create an array of objects
        new = np.ascontiguousarray(['x'] * len(slices) * samples_per_pixel,
                                   dtype=np.dtype(object))

        for i, s in enumerate(slices):
            array = s.memmap()
            if samples_per_pixel == 1:
                assert array.ndim == 2
                new[i] = t(array)

            else:
                for j in range(samples_per_pixel):
                    assert array.ndim == 3
                    new[i * samples_per_pixel + j] = t(array[:, :, j])

        new = np.reshape(new, list(axes.values())[:-2], order='F')
        slices = new

        for x in 'ZCT':
            if x not in ax_keys:
                ax_keys = [x] + ax_keys

        T = ax_keys.index('T')
        C = ax_keys.index('C')
        Z = ax_keys.index('Z')

        while slices.ndim < 3:
            slices = np.expand_dims(slices, 0)

        slices = np.moveaxis(slices, (T,C,Z), (0,1,2))
        return slices

    @classmethod
    def write(cls, images, io=None, big_tiff=True, closefd=True, tags=None, imagej_shape=None):
        '''
        Writes all `images` (which should be a list of 3D (H,W,C) numpy arrays).
        Returns the io object

        If `io` is a string, we write the image to that path,
        if `io` is None, we return a io.BytesIO (io.getbuffer() returns a buffer)
        If `imagej` is a tuple: write this TCZ shape as ImageJ data description
        '''
        if io is None:
            io = Writer(_io.BytesIO())
        elif type(io) is str:
            io = Writer(open(io, 'wb'))
        else:
            io = Writer(io)

        if sys.byteorder == 'little':
            io.write_bytes(b'II')
        else:
            io.write_bytes(b'MM')

        if big_tiff:
            offset_width = 8
            write_offset = io.write_u64
            io.write_u16(43) # magic
            io.write_u16(8) # offset width
            io.write_u16(0)
        else:
            offset_width = 4
            write_offset = io.write_u32
            io.write_u16(42) # magic

        next_ifd_position = len(io) + 8
        write_offset(next_ifd_position)

        Tag = tif_format.TifFormat.Tag
        TagType = tif_format.TifFormat.TagType

        tags = []
        if imagej_shape:
            imagej = '''ImageJ=1.51d
images={}
frames={}
channels={}
slices={}
'''.format(len(images), *imagej_shape).encode('utf-8')
            imagej += b'\x00'
            tags += [(Tag.image_description, TagType.string, imagej)]

        for i, img in enumerate(images):
            io_data = Writer(_io.BytesIO())
            cls._write_image(io, io_data, img, len(io), big_tiff, tags)

            if i == len(images) - 1:
                next_ifd_position = 0
            else:
                next_ifd_position = len(io) + offset_width + len(io_data) + img.nbytes

            write_offset(next_ifd_position)
            io.write_bytes(io_data.get_bytes())

            if type(img) is PlaceHolder:
                io.write_hole(img.nbytes)
            else:
                io.write_bytes(img.tobytes())

        if closefd:
            io.close()

        return io.io


    @classmethod
    def _write_image(cls, io_ifd, io_data, img, total_offset, big_tiff, tags=None):
        Tag = tif_format.TifFormat.Tag
        TagType = tif_format.TifFormat.TagType

        if big_tiff:
            offset_width = 8
            offset_type = TagType.u8
            write_offset = io_ifd.write_u64
        else:
            offset_width = 4
            offset_type = TagType.u4
            write_offset = io_ifd.write_u32

        height, width, channels = img.shape
        kind = {
            'u': 1,
            'i': 2,
            'f': 3,
        }[img.dtype.kind]

        tags = tags or []
        tags += [
                (Tag.image_length, TagType.u4, [height]),
                (Tag.image_width, TagType.u4, [width]),
                (Tag.photometric_interpretation, TagType.u2, [1]),
                (Tag.compression, TagType.u2, [1]),
                (Tag.bits_per_sample, TagType.u2, [img.dtype.itemsize * 8]),
                (Tag.sample_format, TagType.u2, [kind]),
                (Tag.samples_per_pixel, TagType.u2, [channels]),
                (Tag.x_resolution, TagType.u_ratio, [(1, 1)]),
                (Tag.y_resolution, TagType.u_ratio, [(1, 1)]),
                (Tag.strip_byte_counts, TagType.u4, [img.nbytes]),
                (Tag.rows_per_strip, TagType.u4, [height]),
                ]
        tags = sorted(tags, key=lambda t: t[0].value)

        # the len(tags) + 1 is for strip_offsets which we write later
        ifd_len = offset_width + (len(tags) + 1) * (20 if big_tiff else 12) + offset_width
        write_offset(len(tags) + 1)

        for t in tags:
            offset = total_offset + ifd_len + len(io_data)
            io_now = Writer(_io.BytesIO())
            io_later = Writer(_io.BytesIO())
            cls._write_tag(io_now, io_later, t, offset, big_tiff)
            io_ifd.write_bytes(io_now.get_bytes())
            io_data.write_bytes(io_later.get_bytes())

        strip_offsets = [ total_offset + ifd_len + len(io_data)]
        t = (Tag.strip_offsets, offset_type, strip_offsets)
        offset = total_offset + ifd_len + len(io_data)

        cls._write_tag(io_ifd, io_data, t, offset, big_tiff)

        return io_ifd, io_data


    @classmethod
    def _write_tag(cls, io_now, io_later, tag, offset, big_tiff):
        TagType = tif_format.TifFormat.TagType
        if big_tiff:
            offset_width = 8
            offset_type = TagType.u8
            write_offset = io_now.write_u64
        else:
            offset_width = 4
            offset_type = TagType.u4
            write_offset = io_now.write_u32

        kind, typ, values = tag
        old_now_len = len(io_now)

        io_now.write_u16(kind.value)
        io_now.write_u16(typ.value)
        write_offset(len(values))

        itemsize = {
            TagType.u2: 2,
            TagType.u4: 4,
            TagType.u8: 8,
            TagType.u_ratio: 8,
            TagType.s2: 2,
            TagType.s4: 4,
            TagType.s8: 8,
            TagType.s_ratio: 8,
            TagType.string: len(values),
        }[typ]

        inline = itemsize * len(values) <= offset_width

        io = Writer(_io.BytesIO())
        if typ == TagType.string:
            io.write_tag_value(typ, values, offset)
        else:
            for v in values:
                io.write_tag_value(typ, v, offset)

        if inline:
            io_now.write_bytes(io.get_bytes())
            io_now.write_bytes(b'\x00'*(4 + 2*offset_width - len(io_now)))
        else:
            write_offset(offset)
            io_later.write_bytes(io.get_bytes())

        if big_tiff:
            assert old_now_len == len(io_now) - 20
        else:
            assert old_now_len == len(io_now) - 12

        return io_now, io_later


class PlaceHolder(object):
    '''
    A place holder that can be used if you do not really want to
    write any image data (behaves like a numpy array).
    '''
    def __init__(self, shape, dtype):
        self.shape = shape
        self.dtype = np.dtype(dtype)

    @property
    def size(self):
        return np.prod(self.shape)

    @property
    def nbytes(self):
        return self.dtype.itemsize * self.size


if __name__ == '__main__':
    main()



# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
from kaitaistruct import __version__ as ks_version, KaitaiStruct, KaitaiStream, BytesIO
from enum import Enum


if parse_version(ks_version) < parse_version('0.7'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.7 or later is required, but you have %s" % (ks_version))

class TifFormat(KaitaiStruct):

    class Endian(Enum):
        le = 18761
        be = 19789

    class TagType(Enum):
        u1 = 1
        string = 2
        u2 = 3
        u4 = 4
        u_ratio = 5
        s1 = 6
        undefined = 7
        s2 = 8
        s4 = 9
        s_ratio = 10
        f4 = 11
        f8 = 12
        u8 = 16
        s8 = 17
        ifd8 = 18

    class Tag(Enum):
        new_subfile_type = 254
        subfile_type = 255
        image_width = 256
        image_length = 257
        bits_per_sample = 258
        compression = 259
        photometric_interpretation = 262
        threshholding = 263
        cell_width = 264
        cell_length = 265
        fill_order = 266
        document_name = 269
        image_description = 270
        make = 271
        model = 272
        strip_offsets = 273
        orientation = 274
        samples_per_pixel = 277
        rows_per_strip = 278
        strip_byte_counts = 279
        min_sample_value = 280
        max_sample_value = 281
        x_resolution = 282
        y_resolution = 283
        planar_configuration = 284
        page_name = 285
        x_position = 286
        y_position = 287
        free_offsets = 288
        free_byte_counts = 289
        gray_response_unit = 290
        gray_response_curve = 291
        t4_options = 292
        t6_options = 293
        resolution_unit = 296
        page_number = 297
        transfer_function = 301
        software = 305
        date_time = 306
        artist = 315
        host_computer = 316
        predictor = 317
        white_point = 318
        primary_chromaticities = 319
        color_map = 320
        halftone_hints = 321
        tile_width = 322
        tile_length = 323
        tile_offsets = 324
        tile_byte_counts = 325
        bad_fax_lines = 326
        clean_fax_data = 327
        consecutive_bad_fax_lines = 328
        sub_if_ds = 330
        ink_set = 332
        ink_names = 333
        number_of_inks = 334
        dot_range = 336
        target_printer = 337
        extra_samples = 338
        sample_format = 339
        s_min_sample_value = 340
        s_max_sample_value = 341
        transfer_range = 342
        clip_path = 343
        x_clip_path_units = 344
        y_clip_path_units = 345
        indexed = 346
        jpeg_tables = 347
        opi_proxy = 351
        global_parameters_ifd = 400
        profile_type = 401
        fax_profile = 402
        coding_methods = 403
        version_year = 404
        mode_number = 405
        decode = 433
        default_image_color = 434
        jpeg_proc = 512
        jpeg_interchange_format = 513
        jpeg_interchange_format_length = 514
        jpeg_restart_interval = 515
        jpeg_lossless_predictors = 517
        jpeg_point_transforms = 518
        jpegq_tables = 519
        jpegdc_tables = 520
        jpegac_tables = 521
        y_cb_cr_coefficients = 529
        y_cb_cr_sub_sampling = 530
        y_cb_cr_positioning = 531
        reference_black_white = 532
        strip_row_counts = 559
        xmp = 700
        image_id = 32781
        copyright = 33432
        image_layer = 34732
        ij_metadata_byte_counts = 50838
        ij_metadata = 50839
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._read()

    def _read(self):
        self.endian = self._root.Endian(self._io.read_u2be())
        self.header = self._root.Header(self._io, self, self._root)

    class Ifd(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            _on = self._root.endian
            if _on == self._root.Endian.le:
                self._is_le = True
            elif _on == self._root.Endian.be:
                self._is_le = False

            if self._is_le == True:
                self._read_le()
            elif self._is_le == False:
                self._read_be()
            else:
                raise Exception("Unable to decide endianness")

        def _read_le(self):
            _on = self._root.header.is_big_tiff
            if _on == True:
                self.count = self._io.read_u8le()
            elif _on == False:
                self.count = self._io.read_u2le()
            self.entries = [None] * (self.count)
            for i in range(self.count):
                self.entries[i] = self._root.IfdEntry(self._io, self, self._root)

            self.next_ifd_position = self._root.Offset(self._io, self, self._root)

        def _read_be(self):
            _on = self._root.header.is_big_tiff
            if _on == True:
                self.count = self._io.read_u8be()
            elif _on == False:
                self.count = self._io.read_u2be()
            self.entries = [None] * (self.count)
            for i in range(self.count):
                self.entries[i] = self._root.IfdEntry(self._io, self, self._root)

            self.next_ifd_position = self._root.Offset(self._io, self, self._root)

        @property
        def next_ifd(self):
            if hasattr(self, '_m_next_ifd'):
                return self._m_next_ifd if hasattr(self, '_m_next_ifd') else None

            if self.next_ifd_position.value != 0:
                _pos = self._io.pos()
                self._io.seek(self.next_ifd_position.value)
                if self._is_le:
                    self._m_next_ifd = self._root.Ifd(self._io, self, self._root)
                else:
                    self._m_next_ifd = self._root.Ifd(self._io, self, self._root)
                self._io.seek(_pos)

            return self._m_next_ifd if hasattr(self, '_m_next_ifd') else None


    class TagValue(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            _on = self._root.endian
            if _on == self._root.Endian.le:
                self._is_le = True
            elif _on == self._root.Endian.be:
                self._is_le = False

            if self._is_le == True:
                self._read_le()
            elif self._is_le == False:
                self._read_be()
            else:
                raise Exception("Unable to decide endianness")

        def _read_le(self):
            self.array = []
            while not self._io.is_eof():
                _on = self._parent.tag_type
                if _on == self._root.TagType.u_ratio:
                    self.array.append(self._root.URatio(self._io, self, self._root))
                elif _on == self._root.TagType.s2:
                    self.array.append(self._io.read_s2le())
                elif _on == self._root.TagType.s1:
                    self.array.append(self._io.read_s1())
                elif _on == self._root.TagType.u4:
                    self.array.append(self._io.read_u4le())
                elif _on == self._root.TagType.s4:
                    self.array.append(self._io.read_s4le())
                elif _on == self._root.TagType.u1:
                    self.array.append(self._io.read_u1())
                elif _on == self._root.TagType.f4:
                    self.array.append(self._io.read_f4le())
                elif _on == self._root.TagType.u2:
                    self.array.append(self._io.read_u2le())
                elif _on == self._root.TagType.undefined:
                    self.array.append(self._io.read_u1())
                elif _on == self._root.TagType.f8:
                    self.array.append(self._io.read_f8le())
                elif _on == self._root.TagType.s_ratio:
                    self.array.append(self._root.SRatio(self._io, self, self._root))
                elif _on == self._root.TagType.string:
                    self.array.append(self._root.String(self._io, self, self._root))
                else:
                    self.array.append(self._io.read_u8le())


        def _read_be(self):
            self.array = []
            while not self._io.is_eof():
                _on = self._parent.tag_type
                if _on == self._root.TagType.u_ratio:
                    self.array.append(self._root.URatio(self._io, self, self._root))
                elif _on == self._root.TagType.s2:
                    self.array.append(self._io.read_s2be())
                elif _on == self._root.TagType.s1:
                    self.array.append(self._io.read_s1())
                elif _on == self._root.TagType.u4:
                    self.array.append(self._io.read_u4be())
                elif _on == self._root.TagType.s4:
                    self.array.append(self._io.read_s4be())
                elif _on == self._root.TagType.u1:
                    self.array.append(self._io.read_u1())
                elif _on == self._root.TagType.f4:
                    self.array.append(self._io.read_f4be())
                elif _on == self._root.TagType.u2:
                    self.array.append(self._io.read_u2be())
                elif _on == self._root.TagType.undefined:
                    self.array.append(self._io.read_u1())
                elif _on == self._root.TagType.f8:
                    self.array.append(self._io.read_f8be())
                elif _on == self._root.TagType.s_ratio:
                    self.array.append(self._root.SRatio(self._io, self, self._root))
                elif _on == self._root.TagType.string:
                    self.array.append(self._root.String(self._io, self, self._root))
                else:
                    self.array.append(self._io.read_u8be())



    class String(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.string = (self._io.read_bytes_term(0, False, True, True)).decode(u"utf-8")


    class Offset(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            _on = self._root.endian
            if _on == self._root.Endian.le:
                self._is_le = True
            elif _on == self._root.Endian.be:
                self._is_le = False

            if self._is_le == True:
                self._read_le()
            elif self._is_le == False:
                self._read_be()
            else:
                raise Exception("Unable to decide endianness")

        def _read_le(self):
            _on = self._root.header.is_big_tiff
            if _on == False:
                self.value = self._io.read_u4le()
            elif _on == True:
                self.value = self._io.read_u8le()

        def _read_be(self):
            _on = self._root.header.is_big_tiff
            if _on == False:
                self.value = self._io.read_u4be()
            elif _on == True:
                self.value = self._io.read_u8be()


    class Header(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            _on = self._root.endian
            if _on == self._root.Endian.le:
                self._is_le = True
            elif _on == self._root.Endian.be:
                self._is_le = False

            if self._is_le == True:
                self._read_le()
            elif self._is_le == False:
                self._read_be()
            else:
                raise Exception("Unable to decide endianness")

        def _read_le(self):
            self.magic = self._io.read_u2le()
            if self.is_big_tiff:
                self.offset_size_raw = self._io.read_u2le()

            if not (self.is_big_tiff):
                self.first_ifd_position_u4 = self._io.read_u4le()

            if self.is_big_tiff:
                self.zeros = self._io.read_u2le()

            if self.is_big_tiff:
                self.first_ifd_position_big_tiff = self._io.read_u8le()


        def _read_be(self):
            self.magic = self._io.read_u2be()
            if self.is_big_tiff:
                self.offset_size_raw = self._io.read_u2be()

            if not (self.is_big_tiff):
                self.first_ifd_position_u4 = self._io.read_u4be()

            if self.is_big_tiff:
                self.zeros = self._io.read_u2be()

            if self.is_big_tiff:
                self.first_ifd_position_big_tiff = self._io.read_u8be()


        @property
        def is_big_tiff(self):
            if hasattr(self, '_m_is_big_tiff'):
                return self._m_is_big_tiff if hasattr(self, '_m_is_big_tiff') else None

            self._m_is_big_tiff = self.magic == 43
            return self._m_is_big_tiff if hasattr(self, '_m_is_big_tiff') else None

        @property
        def first_ifd(self):
            if hasattr(self, '_m_first_ifd'):
                return self._m_first_ifd if hasattr(self, '_m_first_ifd') else None

            if not (self.is_big_tiff):
                _pos = self._io.pos()
                self._io.seek(self.first_ifd_position_u4)
                if self._is_le:
                    self._m_first_ifd = self._root.Ifd(self._io, self, self._root)
                else:
                    self._m_first_ifd = self._root.Ifd(self._io, self, self._root)
                self._io.seek(_pos)

            return self._m_first_ifd if hasattr(self, '_m_first_ifd') else None

        @property
        def first_ifd_big_tiff(self):
            if hasattr(self, '_m_first_ifd_big_tiff'):
                return self._m_first_ifd_big_tiff if hasattr(self, '_m_first_ifd_big_tiff') else None

            if self.is_big_tiff:
                _pos = self._io.pos()
                self._io.seek(self.first_ifd_position_big_tiff)
                if self._is_le:
                    self._m_first_ifd_big_tiff = self._root.Ifd(self._io, self, self._root)
                else:
                    self._m_first_ifd_big_tiff = self._root.Ifd(self._io, self, self._root)
                self._io.seek(_pos)

            return self._m_first_ifd_big_tiff if hasattr(self, '_m_first_ifd_big_tiff') else None


    class SRatio(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            _on = self._root.endian
            if _on == self._root.Endian.le:
                self._is_le = True
            elif _on == self._root.Endian.be:
                self._is_le = False

            if self._is_le == True:
                self._read_le()
            elif self._is_le == False:
                self._read_be()
            else:
                raise Exception("Unable to decide endianness")

        def _read_le(self):
            self.nominator = self._io.read_s4le()
            self.denominator = self._io.read_s4le()

        def _read_be(self):
            self.nominator = self._io.read_s4be()
            self.denominator = self._io.read_s4be()

        @property
        def value(self):
            if hasattr(self, '_m_value'):
                return self._m_value if hasattr(self, '_m_value') else None

            self._m_value = ((1.0 * self.nominator) / self.denominator)
            return self._m_value if hasattr(self, '_m_value') else None


    class IfdEntry(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            _on = self._root.endian
            if _on == self._root.Endian.le:
                self._is_le = True
            elif _on == self._root.Endian.be:
                self._is_le = False

            if self._is_le == True:
                self._read_le()
            elif self._is_le == False:
                self._read_be()
            else:
                raise Exception("Unable to decide endianness")

        def _read_le(self):
            self.tag_raw = self._io.read_u2le()
            self.tag_type_raw = self._io.read_u2le()
            _on = self._root.header.is_big_tiff
            if _on == False:
                self.count = self._io.read_u4le()
            elif _on == True:
                self.count = self._io.read_u8le()
            if self.is_value:
                self._raw_values = self._io.read_bytes((self.tag_type_length * self.count))
                io = KaitaiStream(BytesIO(self._raw_values))
                self.values = self._root.TagValue(io, self, self._root)

            if self.is_value:
                self.padding = self._io.read_bytes((self.value_size - (self.tag_type_length * self.count)))

            if not (self.is_value):
                self.offset = self._root.Offset(self._io, self, self._root)


        def _read_be(self):
            self.tag_raw = self._io.read_u2be()
            self.tag_type_raw = self._io.read_u2be()
            _on = self._root.header.is_big_tiff
            if _on == False:
                self.count = self._io.read_u4be()
            elif _on == True:
                self.count = self._io.read_u8be()
            if self.is_value:
                self._raw_values = self._io.read_bytes((self.tag_type_length * self.count))
                io = KaitaiStream(BytesIO(self._raw_values))
                self.values = self._root.TagValue(io, self, self._root)

            if self.is_value:
                self.padding = self._io.read_bytes((self.value_size - (self.tag_type_length * self.count)))

            if not (self.is_value):
                self.offset = self._root.Offset(self._io, self, self._root)


        @property
        def is_value(self):
            if hasattr(self, '_m_is_value'):
                return self._m_is_value if hasattr(self, '_m_is_value') else None

            self._m_is_value = (self.tag_type_length * self.count) <= self.value_size
            return self._m_is_value if hasattr(self, '_m_is_value') else None

        @property
        def value_size(self):
            if hasattr(self, '_m_value_size'):
                return self._m_value_size if hasattr(self, '_m_value_size') else None

            self._m_value_size = (4 + ((int(self._root.header.is_big_tiff) & 1) * 4))
            return self._m_value_size if hasattr(self, '_m_value_size') else None

        @property
        def tag(self):
            if hasattr(self, '_m_tag'):
                return self._m_tag if hasattr(self, '_m_tag') else None

            self._m_tag = self._root.Tag(self.tag_raw)
            return self._m_tag if hasattr(self, '_m_tag') else None

        @property
        def tag_type_length(self):
            if hasattr(self, '_m_tag_type_length'):
                return self._m_tag_type_length if hasattr(self, '_m_tag_type_length') else None

            self._m_tag_type_length = (1 if self.tag_type == self._root.TagType.u1 else (1 if self.tag_type == self._root.TagType.string else (2 if self.tag_type == self._root.TagType.u2 else (4 if self.tag_type == self._root.TagType.u4 else (8 if self.tag_type == self._root.TagType.u8 else (8 if self.tag_type == self._root.TagType.u_ratio else (1 if self.tag_type == self._root.TagType.s1 else (1 if self.tag_type == self._root.TagType.undefined else (2 if self.tag_type == self._root.TagType.s2 else (4 if self.tag_type == self._root.TagType.s4 else (8 if self.tag_type == self._root.TagType.s8 else (8 if self.tag_type == self._root.TagType.s_ratio else (4 if self.tag_type == self._root.TagType.f4 else (8 if self.tag_type == self._root.TagType.f8 else 999999999))))))))))))))
            return self._m_tag_type_length if hasattr(self, '_m_tag_type_length') else None

        @property
        def tag_type(self):
            if hasattr(self, '_m_tag_type'):
                return self._m_tag_type if hasattr(self, '_m_tag_type') else None

            self._m_tag_type = self._root.TagType(self.tag_type_raw)
            return self._m_tag_type if hasattr(self, '_m_tag_type') else None

        @property
        def external_values(self):
            if hasattr(self, '_m_external_values'):
                return self._m_external_values if hasattr(self, '_m_external_values') else None

            _pos = self._io.pos()
            self._io.seek(self.offset.value)
            if self._is_le:
                self._raw__m_external_values = self._io.read_bytes((self.tag_type_length * self.count))
                io = KaitaiStream(BytesIO(self._raw__m_external_values))
                self._m_external_values = self._root.TagValue(io, self, self._root)
            else:
                self._raw__m_external_values = self._io.read_bytes((self.tag_type_length * self.count))
                io = KaitaiStream(BytesIO(self._raw__m_external_values))
                self._m_external_values = self._root.TagValue(io, self, self._root)
            self._io.seek(_pos)
            return self._m_external_values if hasattr(self, '_m_external_values') else None


    class URatio(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            _on = self._root.endian
            if _on == self._root.Endian.le:
                self._is_le = True
            elif _on == self._root.Endian.be:
                self._is_le = False

            if self._is_le == True:
                self._read_le()
            elif self._is_le == False:
                self._read_be()
            else:
                raise Exception("Unable to decide endianness")

        def _read_le(self):
            self.nominator = self._io.read_u4le()
            self.denominator = self._io.read_u4le()

        def _read_be(self):
            self.nominator = self._io.read_u4be()
            self.denominator = self._io.read_u4be()

        @property
        def value(self):
            if hasattr(self, '_m_value'):
                return self._m_value if hasattr(self, '_m_value') else None

            self._m_value = ((1.0 * self.nominator) / self.denominator)
            return self._m_value if hasattr(self, '_m_value') else None






from setuptools import setup

def readme():
    with open('README.md') as f:
        return f.read()

setup(
    name='bigtiff',
    version='0.1.1',
    description='A library for random access for large (Big-)TIFF files',
    author='Manuel Schölling',
    author_email='manuel.schoelling@dzne.de',
    license='GPL3',
    packages=['bigtiff'],
    zip_safe=False,
    install_requires=[
        'numpy',
        'kaitaistruct==0.8.0'
    ],
)



import ctypes.util
from mmap import MAP_SHARED
import os

INVALID_POINTER = ctypes.c_void_p(-1).value

def check_mmap(addr, func, args):
    errno = ctypes.get_errno()
    err = os.strerror(errno)
    assert addr != INVALID_POINTER, 'mmap() failed: {}'.format(err)
    return addr


def check_munmap(ret, func, args):
    errno = ctypes.get_errno()
    assert ret != -1, 'munmap() failed: {}'.format(os.strerror(errno))


libc = ctypes.util.find_library('c')
libc = ctypes.CDLL(libc)

libc.mmap.argtypes = (ctypes.c_void_p, ctypes.c_size_t, ctypes.c_int,
                      ctypes.c_int, ctypes.c_int, ctypes.c_size_t)
libc.mmap.restype = ctypes.c_void_p
libc.mmap.errcheck = check_mmap

libc.munmap.argtypes = (ctypes.c_void_p, ctypes.c_size_t)
libc.munmap.restype = ctypes.c_int
libc.munmap.errcheck = check_munmap

from numpy.core import memmap as np_memmap
import unittest.mock

def memmap(*args, **kwds):
    '''
    Works just like the ordinary numpy.memmap() but does not
    open an additional file descriptor.
    '''
    with unittest.mock.patch('mmap.mmap', new=mmap):
        arr = np_memmap(*args, **kwds)
    return arr


def mmap(fileno, bytelen, access=None, offset=0):
    '''
    Works just like the ordinary mmap.mmap() but does not
    open an additional file descriptor.
    '''
    buffer_type = ctypes.c_char * bytelen
    class MMap(buffer_type):
        def __del__(self):
            libc.munmap(addr, bytelen)

    addr = libc.mmap(None, bytelen, access, MAP_SHARED, fileno, offset)
    return MMap.from_address(addr)

